#!/usr/bin/env python
# encoding: utf-8
"""
Author: Nick Bearson, nickb@ssec.wisc.edu
Copyright (c) 2012 University of Wisconsin SSEC. All rights reserved.
"""

import numpy as np

from datetime import datetime
import glob
import re
import os

from descriptors import reify
import grain

import viirsmend.viirsmend as vm

import logging
log = logging.getLogger(__name__)


VERSION = "mjh fixed HIRS bands 2014-07-21; mjh fixed float types 2014-07-22; mjh working collocation 2014-07-24; mjh fixed AVHRR fills + divide refls by 100 + fill in SensorAzimuth; mjh fill in AVHRR EmissiveBands with rads converted from brightness temps; mjh fix HIRS radiance units"


def group_files(filelist, sensor):
  geo_files = []

  for f in filelist:
    m = re.match(sensor.GROUPING_REGEX, f)
    if m:
      geo_files.append(f)
    else:
      continue

  group_list = []
  for geo in geo_files:  # FIXME: clunky atm, want to match geo files with all files that share their times
    geo_start, geo_end = sensor.date_bounds_from_fname(geo)
    geo_group = []

    for f in filelist:
      f_start, f_end = sensor.date_bounds_from_fname(f)
      if f_start <= geo_end and f_end >= geo_start:
        geo_group.append(f)

    if len(geo_group) > 0:
      group_list.append(geo_group)

  return group_list



class Aggregator(object):

  def __init__(self, platform_name, sensor, tag, file_list, start_datetime, end_datetime, fix_bowtie=False):
    self._file_list = file_list
    self._start_datetime = start_datetime
    self._end_datetime = end_datetime
    self._fix_bowtie = fix_bowtie
    self._platform_name = platform_name
    self._sensor = sensor
    self._tag = tag

    self._sensor_list = []
    
    file_groups = group_files(file_list, sensor)

    for group in file_groups:
      self._sensor_list.append(sensor(platform=platform_name, tag=tag, file_list=group))

    if len(self._sensor_list) is 0:
      raise RuntimeError, "No sensors could be created."

    self.nDETECTORS = self._sensor_list[0].nDETECTORS

    self.grain = grain.Grain()

    s_tai = self.grain.utc2tai(start_datetime, epoch=sensor.SENSOR_EPOCH)
    e_tai = self.grain.utc2tai(end_datetime,   epoch=sensor.SENSOR_EPOCH)
    log.debug("BOUNDING TIMES | %s - %s" % (s_tai, e_tai))
    self._get_keys(s_tai, e_tai)



  @reify
  def PLATFORM(self):
    return self._sensor_list[0].PLATFORM

  @reify
  def SENSOR(self):
    return self._sensor_list[0].SENSOR

  @reify
  def SOFTWARE_VERSION(self):
    return VERSION

  @reify
  def FILENAMES(self):
    filenames = []
    for f in self._file_list:
      filenames.append(os.path.basename(f))
    filenames = ", ".join(filenames)  # FIXME: TURNED ARRAY INTO STRING LIST FOR HDF4
    return filenames

  @reify
  def CREATION_DATE(self):
    return str(datetime.utcnow())

  @reify
  def START_TIME(self):
    return str(self._start_datetime)

  @reify
  def END_TIME(self):
    return str(self._end_datetime)

  @reify
  def CREATION_HOST(self):
    from socket import gethostname
    return gethostname()


  def _get_keys(self, min, max):
    scan_times = self._agg_op('ScanStartTime', 1)
    row_times = np.repeat(scan_times, self.nDETECTORS)
    det_array = np.array(range(self.nDETECTORS))
    sort_offsets = np.tile(det_array, scan_times.shape)
    sort_keys = (sort_offsets, row_times)
    self._sort_key = np.lexsort(sort_keys)
    log.debug("SORT KEY:")
    log.debug(self._sort_key)
    sorted = row_times[self._sort_key]
    log.debug("SORTED:")
    log.debug(sorted)
    self._bounds_key = (sorted >= min) & (sorted < (max+1))  # this is so we include the ENTIRE second
    if np.alltrue(self._bounds_key == False):
      raise RuntimeError, "No scan lines fit within the bounds specified!"
    log.debug("BOUNDS KEY")
    log.debug(self._bounds_key)
    log.debug(min)
    log.debug(max)


  # general aggregation operation for N dimension arrays
  def _agg_op(self, funcname, n_dims, *args):
    datalist = []
    for sensor in self._sensor_list:
      f = getattr(sensor, funcname)
      if hasattr(f, '__call__'):
        log.debug("aggregator calling %s" % (f))
        f = f(*args)
        log.debug("aggregator returned")
      datalist.append(f)
    # If "None" gets returned for everything, keep passing it through:
    if all(v is None for v in datalist):
      return None
    log.debug("agg_op combining data from " + funcname)
    unsorted = np.concatenate(tuple(datalist), axis=(n_dims-2))
    return unsorted


  # general merge operation for singular dimension arrays where they all should be the same and we just want one
  def _merge_op(self, funcname):
    datalist = []
    for sensor in self._sensor_list:
      f = getattr(sensor, funcname)
      datalist.append(f)
    try:
      def sliding_window(seq, size=2):
        for i in xrange(len(seq) - size + 1):
          yield seq[i:i+size]

      for a, b in sliding_window(datalist, 2):
        assert(np.all(a == b))

    except:
      log.error("%s trying to merge unequal arrays:" % (funcname))
      log.error(datalist)
      log.error("EXITING!")
      exit(1)
    return datalist[0]


  @reify
  def _ViirsMender(self):
    lats = self.Latitude
    lons = self.Longitude

    if self._tag == "svm":
      res = vm.MOD_RESOLUTION
    elif self._tag == "svi":
      res = vm.IMG_RESOLUTION
    else:
      raise NotImplementedError

    return vm.ViirsMender(lons, lats, res)

  def _apply_bowtie_fix(self, bands): # TODO: make this less viirs-centric?
    log.warn("using new viirs mender!")
    vmer = self._ViirsMender
    for i in range(len(bands)):
      vmer.mend(bands[i])


  @reify
  def ReflectiveSolarBands(self):
    rsb = self._agg_op('ReflectiveSolarBands', 3)
    rsb = rsb[:,self._sort_key][:,self._bounds_key]
    if self._fix_bowtie:
      self._apply_bowtie_fix(rsb)
    return rsb

  @reify
  def ReflectiveBandNums(self):
    nums = self._merge_op('ReflectiveBandNums')
    return ", ".join(nums)

  @reify
  def ReflectiveBandCenters(self):
    centers = self._merge_op('ReflectiveBandCenters')
    return centers

  @reify
  def EmissiveBands(self):
    eb = self._agg_op('EmissiveBands', 3)
    eb = eb[:,self._sort_key][:,self._bounds_key]
    if self._fix_bowtie:
      self._apply_bowtie_fix(eb)

    cb = self._agg_op('SounderEmissive', 3)
    cb = cb[:,self._sort_key][:,self._bounds_key]

    combined = np.vstack((eb,cb)).astype(np.float32)
    log.debug("EmissiveBands shape: %s, dtype: %s" % (combined.shape, combined.dtype))
    return combined

  @reify
  def EmissiveBandNums(self):
    nums = self._merge_op('EmissiveBandNums')
    nums = ", ".join(nums)
    cnums = self._merge_op('SounderNums')
    cnums = ", ".join(map(str, cnums))
    combined = nums + ", " + cnums
    return combined

  @reify
  def EmissiveBandCenters(self):
    centers = self._merge_op('EmissiveBandCenters')
    ccenters = self._merge_op('SounderCenters')
    ccenters = np.array(ccenters)  # FIXME: should this be turned into a numpy array somewhere else?
    centers = np.concatenate((centers, ccenters)).astype(np.float32)
    log.debug("EmissiveBandCenters shape: %s, dtype: %s" % (centers.shape, centers.dtype))
    return centers

  @reify
  def BrightnessTemperatureBands(self):
    eb = self._agg_op('BrightnessTemperatureBands', 3)
    eb = eb[:,self._sort_key][:,self._bounds_key]
    if self._fix_bowtie:
      self._apply_bowtie_fix(eb)

    cb = self._agg_op('SounderBTs', 3)
    cb = cb[:,self._sort_key][:,self._bounds_key]

    combined = np.vstack((eb,cb)).astype(np.float32)
    log.debug("BrightnessTemperatureBands shape: %s, dtype: %s" % (combined.shape, combined.dtype))
    return combined

  @reify
  def BrightnessTemperatureBandNums(self):
    nums = self._merge_op('BrightnessTemperatureBandNums')
    nums = ", ".join(nums)
    cnums = self._merge_op('SounderNums')
    cnums = ", ".join(map(str, cnums))
    combined = nums + ", " + cnums
    return combined

  @reify
  def BrightnessTemperatureBandCenters(self):
    centers = self._merge_op('BrightnessTemperatureBandCenters')
    ccenters = self._merge_op('SounderCenters')
    ccenters = np.array(ccenters)  # FIXME: should this be turned into a numpy array somewhere else?
    centers = np.concatenate((centers, ccenters)).astype(np.float32)
    log.debug("EmissiveBandCenters shape: %s, dtype: %s" % (centers.shape, centers.dtype))
    return centers

  @reify
  def Height(self):
    h = self._agg_op('Height', 2)
    return h[self._sort_key][self._bounds_key]
    
  @reify
  def SensorZenith(self):
    sz = self._agg_op('SensorZenith', 2)
    return sz[self._sort_key][self._bounds_key]

  @reify
  def SensorAzimuth(self):
    sa = self._agg_op('SensorAzimuth', 2)
    return sa[self._sort_key][self._bounds_key]

  @reify
  def Range(self):
    r = self._agg_op('Range', 2)
    return r[self._sort_key][self._bounds_key]
  
  @reify
  def SolarZenith(self):
    sz = self._agg_op('SolarZenith', 2)
    return sz[self._sort_key][self._bounds_key]
  
  @reify
  def SolarAzimuth(self):
    sa = self._agg_op('SolarAzimuth', 2)
    return sa[self._sort_key][self._bounds_key]
  
  @reify
  def Latitude(self):
    lat = self._agg_op('Latitude', 2)
    return lat[self._sort_key][self._bounds_key]
    
  @reify
  def Longitude(self):
    lon = self._agg_op('Longitude', 2)
    return lon[self._sort_key][self._bounds_key]
  
  @reify
  def ScanStartTime(self):
    scan_times = self._agg_op('ScanStartTime', 1)
    row_times = np.repeat(scan_times, self.nDETECTORS)
    return row_times[self._sort_key][self._bounds_key]

  @reify
  def LandSeaMask(self):
    lm =  self._agg_op('LandSeaMask', 2)
    return lm[self._sort_key][self._bounds_key]

  @reify
  def SounderElementIndex(self):
    ind =  self._agg_op('SounderElementIndex', 2)
    return ind[self._sort_key][self._bounds_key]

  @reify
  def SounderLineIndex(self):
    ind =  self._agg_op('SounderLineIndex', 2)
    return ind[self._sort_key][self._bounds_key]

  @reify
  def SounderCloudHeights(self):
    ind =  self._agg_op('SounderCloudHeights', 2)
    return ind[self._sort_key][self._bounds_key]

  @reify
  def SounderCloudPressure(self):
    ind =  self._agg_op('SounderCloudPressure', 2)
    return ind[self._sort_key][self._bounds_key]

  @reify
  def SounderCloudTemperature(self):
    ind =  self._agg_op('SounderCloudTemperature', 2)
    return ind[self._sort_key][self._bounds_key]

  @reify
  def HirsMask(self):
    ind =  self._agg_op('HirsMask', 2)
    return ind[self._sort_key][self._bounds_key]


def write_agg_IFF_h4(oAgg, out_name):
  import pyhdf.SD as h4

  dtype_dict = {
           np.int8    : h4.SDC.INT8   ,
           np.uint8   : h4.SDC.UINT8  ,
           np.int16   : h4.SDC.INT16  ,
           np.uint16  : h4.SDC.UINT16 ,
           np.int32   : h4.SDC.INT32  ,
           np.uint32  : h4.SDC.UINT32 ,
           np.float32 : h4.SDC.FLOAT32,
           np.float64 : h4.SDC.FLOAT64,
           }

  if oAgg._tag == "cma":
    from filestruct import get_cm_struct as fstruct
    from filestruct import get_iff_attrs as fattrs
  else:
    from filestruct import get_iff_struct as fstruct
    from filestruct import get_iff_attrs  as fattrs

  if os.path.exists(out_name):
    os.remove(out_name)

  h4out = h4.SD(out_name, h4.SDC.WRITE|h4.SDC.CREATE)

  ga_struct = fattrs(oAgg)
  for ga_name in sorted(ga_struct.keys()):
    log.info("Adding Attribute: %s" % (ga_name))
    setattr(h4out, ga_name, ga_struct[ga_name])

  ds_struct = fstruct(oAgg)

  for ds_name in sorted(ds_struct.keys()):
    log.info("Adding Dataset: %s" % (ds_name))
    f = ds_struct[ds_name]['source_function']
    data = f
    if data is None:  # FIXME: THIS IS SO LANDSEAMASK DOESNT MESS EVERYTHING UP
      continue

    log.info("Writing %s | np type: %s, h4 type %s" % (ds_name, f.dtype.type, dtype_dict[f.dtype.type]) )
    ds = h4out.create(ds_name, dtype_dict[f.dtype.type], f.shape)
    # mjh 2014/10/15 temp. hacking at compression level
    ds.setcompress(h4.SDC.COMP_DEFLATE, 4)
    #ds.setcompress(h4.SDC.COMP_NONE)

    dimnames = ds_struct[ds_name]['dims']
    for i in range(len(dimnames)):
      dim = ds.dim(i)
      log.debug("writing dimension %d: %s" %(i, dimnames[i]))
      dim.setname(dimnames[i])

    ds[:] = data

    for attr_name in ds_struct[ds_name]['attrs'].keys():
      attr = ds_struct[ds_name]['attrs'][attr_name]
      if callable(attr):
        attr = attr()
      setattr(ds, attr_name, attr)
    ds.endaccess()

  h4out.end()




def write_agg_IFF_h5(oAgg, out_name):
  import h5py
  from filestruct import get_iff_struct, get_iff_attrs

  if os.path.exists(out_name):
    os.remove(out_name)

  h5out = h5py.File(out_name, 'w')

  ga_struct = get_iff_attrs(oAgg)
  for ga_name in ga_struct.keys():
    log.info("Adding Attribute: %s" % (ga_name))
    h5out.attrs[ga_name] = ga_struct[ga_name]

  ds_struct = get_iff_struct(oAgg)

  for ds_name in ds_struct.keys():
    log.info("Adding Dataset: %s" % (ds_name))
    f = ds_struct[ds_name]['source_function']
#    print type(f)
    data = f
    if data is None:  # FIXME: THIS IS SO LANDSEAMASK DOESNT MESS EVERYTHING UP
      continue
    ds = h5out.create_dataset(ds_name, data=data, compression='gzip', compression_opts=4)
#    ds = h5out.create_dataset(ds_name, data=data)

    for attr_name in ds_struct[ds_name]['attrs'].keys():
      attr = ds_struct[ds_name]['attrs'][attr_name]
#      print type(attr)
      if callable(attr):
#        print "attribute is a function: ", attr
        attr = attr()
      ds.attrs[attr_name] = attr

  h5out.close()


# FOR DEBUGGING PURPOSES ONLY:
def _enter_debugger(type, value, tb):
  import traceback
  lines=traceback.format_exception(type, value, tb)
  log.debug("---------------------Traceback lines-----------------------")
  log.debug("\n".join(lines))
  log.debug("-----------------------------------------------------------")
  import pdb
  pdb.pm()


if __name__ == "__main__":
  import os
  import sys
#  sys.excepthook = _enter_debugger

  levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
  logging.basicConfig(level = levels[3])  # THIS IS CURRENTLY JUST A TEST ENV SO HARDCODE TO DEBUG!

  try:
    arg = sys.argv[1]
    if arg == 'm':
      test_MODIS()
    elif arg == "v":
      test_VIIRS()
    else:
      print 'argv[1] doesnt match expected'
  except:
    test_VIIRS()
    test_MODIS()

